/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
#ifdef FORCE_NOMPI
#define EXCHANGE_AX {}
#define EXCHANGE_NORME {}
#define IF_RANK_0 if (1)
#else
#define EXCHANGE_AX							\
  {									\
    pastix_float_t * ax_rcv;						\
    ax_rcv = malloc(globn*sizeof(pastix_float_t));			\
    MPI_Allreduce(ax, ax_rcv, globn, MPI_PASTIX_FLOAT, MPI_SUM,		\
		  MPI_COMM_WORLD);					\
    free(ax);								\
    ax = ax_rcv;							\
  }

#define EXCHANGE_NORME							\
  {									\
    pastix_float_t norme1_rcv, norme2_rcv;				\
    MPI_Allreduce(&norme1, &norme1_rcv, 1, MPI_PASTIX_FLOAT,		\
		  MPI_SUM, MPI_COMM_WORLD);				\
    norme1 = norme1_rcv;						\
    MPI_Allreduce(&norme2, &norme2_rcv, 1, MPI_PASTIX_FLOAT,		\
		  MPI_SUM, MPI_COMM_WORLD);				\
    norme2 = norme2_rcv;						\
  }									
#define IF_RANK_0  if (rank == 0)
#endif


#define CHECK_DIST_SOL(colptr2, rows2, values2, rhs2, ncol2, loc2glob2, globn, rhssaved_g) \
  {									\
									\
    pastix_int_t   * glob2loc;						\
    pastix_float_t * ax;						\
    pastix_float_t   norme1, norme2;					\
    pastix_int_t     j, iter;						\
    pastix_float_t * sol_g, *sol_g_recv;				\
    glob2loc = malloc(globn*sizeof(pastix_int_t));			\
    for (iter = 0; iter < globn; iter++)				\
      glob2loc[iter] = -1;						\
    for (iter = 0; iter < ncol2; iter++)				\
      glob2loc[loc2glob2[iter]-1] = iter+1;				\
									\
    sol_g = malloc(globn*sizeof(pastix_float_t));			\
    memset(sol_g, 0, globn*sizeof(pastix_float_t));			\
    sol_g_recv = malloc(globn*sizeof(pastix_float_t));			\
    for (iter= 0; iter < ncol2; iter++)					\
      {									\
	sol_g[loc2glob2[iter]-1] = rhs2[iter];				\
      }									\
    MPI_Allreduce(sol_g, sol_g_recv, globn,				\
		  MPI_PASTIX_FLOAT, MPI_SUM,				\
		  MPI_COMM_WORLD);					\
    free(sol_g);							\
    sol_g =sol_g_recv;							\
									\
    ax = malloc(globn*sizeof(pastix_float_t));				\
    memset(ax, 0, globn*sizeof(pastix_float_t));			\
    for (iter= 0; iter < ncol2; iter++)					\
      {									\
	for (j = colptr2[iter]-1; j < colptr2[iter+1] - 1; j++)		\
	  {								\
	    ax[rows2[j]-1] += values2[j]*sol_g[loc2glob2[iter]-1];	\
	    if ((MTX_ISSYM(type) == 1) &&				\
		(loc2glob2[iter]-1 != (rows2[j]-1)))			\
	      {								\
		ax[loc2glob2[iter]-1] += values2[j]*			\
		  sol_g[rows2[j]-1];					\
	      }								\
	  }								\
      }									\
    free(sol_g);							\
    EXCHANGE_AX;							\
    norme1= norme2 = 0;							\
    for (iter= 0; iter < ncol2; iter++)					\
      {									\
	norme1 += (double)((    ax[loc2glob2[iter]-1] -			\
				rhssaved_g[loc2glob2[iter]-1])*		\
			   conj(ax[loc2glob2[iter]-1] -			\
				rhssaved_g[loc2glob2[iter]-1]));	\
	norme2 += (double)((rhssaved_g[loc2glob2[iter]-1])*		\
			   conj(rhssaved_g[loc2glob2[iter]-1]));	\
      }									\
    EXCHANGE_NORME;							\
    IF_RANK_0 {								\
      fprintf(stdout, "Precision : ||ax -b||/||b|| = %.20lg\n",		\
	      sqrt(norme1/norme2));					\
    }									\
    free(ax);								\
    free(glob2loc);							\
  }
