/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
#ifndef X_INCLUDE_ESSL
#define CBLAS
#endif /* X_INCLUDE_ESSL */

#ifdef BLAS_DOUBLE
#ifdef CPLX
#ifdef CBLAS
#define AXPY cblas_zaxpy
#define GEMM cblas_zgemm
#define GEMV cblas_zgemv
#define TRSM cblas_ztrsm
#define TRSV cblas_ztrsv
#define GER cblas_zgeru
#define SYR cblas_zher
#define SYRK cblas_zherk
#define COPY cblas_zcopy
#define SCAL cblas_zscal
#else /* CBLAS */
#define AXPY zaxpy
#define GEMM zgemm
#define GEMV zgemv
#define TRSM ztrsm
#define TRSV ztrsv
#define POF zpof
#define PPF zppf
#define GER zgeru
#define SYR zher
#define SYRK zherk
#define COPY zcopy
#define SCAL zscal
#endif /* CBLAS */
#else /* CPLX */
#ifdef CBLAS
#define AXPY cblas_daxpy
#define GEMM cblas_dgemm
#define GEMV cblas_dgemv
#define TRSM cblas_dtrsm
#define TRSV cblas_dtrsv
#define GER cblas_dger
#define SYR cblas_dsyr
#define SYRK cblas_dsyrk
#define COPY cblas_dcopy
#define SCAL cblas_dscal
#else /* CBLAS */
#define AXPY                        daxpy
#define GEMM                        dgemm
#define GEMV                        dgemv
#define GEAM                        dgeam
#define GECP                        dgecp
#define SYMM                        gsymm
#define TRSM                        dtrsm
#define TRSV                        dtrsv
#define GEF                         dgef
#define POF                         dpof
#define PPF                         dppf
#define GER                         dger
#define GETMO                       dgetmo
#define SYR                         dsyr
#define SPR                         dspr
#define SYRK                        dsyrk
#define COPY                        dcopy
#define SCAL                        dscal
#define V_DIV                       dv_div
#define POTRF                       dpotrf
#define POTF2                       dpotf2
#endif /* CBLAS */
#endif /* CPLX */
#else /* BLAS_DOUBLE */
#ifdef CPLX
#ifdef CBLAS
#define AXPY cblas_caxpy
#define GEMM cblas_cgemm
#define GEMV cblas_cgemv
#define TRSM cblas_ctrsm
#define TRSV cblas_ctrsv
#define GER cblas_cgeru
#define SYR cblas_cher
#define SYRK cblas_cherk
#define COPY cblas_ccopy
#define SCAL cblas_cscal
#else /* CBLAS */
#define AXPY caxpy
#define GEMM cgemm
#define GEMV cgemv
#define TRSM ctrsm
#define TRSV ctrsv
#define GER cgeru
#define SYR cher
#define SYRK cherk
#define COPY ccopy
#define SCAL cscal
#endif /* CBLAS */
#else /* CPLX */
#ifdef CBLAS
#define AXPY cblas_saxpy
#define GEMM cblas_sgemm
#define GEMV cblas_sgemv
#define TRSM cblas_strsm
#define TRSV cblas_strsv
#define GER cblas_sger
#define SYR cblas_ssyr
#define SYRK cblas_ssyrk
#define COPY cblas_scopy
#define SCAL cblas_sscal
#else /* CBLAS */
#define AXPY                        saxpy
#define GEMM                        sgemm
#define GEMV                        sgemv
#define GEAM                        sgeam
#define GECP                        sgecp
#define SYMM                        ssymm
#define TRSM                        strsm
#define TRSV                        strsv
#define GEF                         sgef
#define POF                         spof
#define PPF                         sppf
#define GER                         sger
#define GETMO                       sgetmo
#define SYR                         ssyr
#define SPR                         sspr
#define SYRK                        ssyrk
#define COPY                        scopy
#define SCAL                        sscal
#define V_DIV                       sv_div
#define POTRF                       spotrf
#define POTF2                       spotf2
#endif /* CBLAS */
#endif /* CPLX */
#endif

#ifdef X_INCLUDE_ESSL
#define IBM_GEAM
/*
#define IBM_GEAM
#define MY_GEAM
*/
#endif /* X_INCLUDE_ESSL */

#ifdef IBM_GEAM
#ifdef BLAS_DOUBLE
#ifdef CPLX
#define GESUB zgesub
#define GEADD zgeadd
#else
#define GEADD dgeadd
#define GESUB dgesub
#endif /* CPLX */
#else
#ifdef CPLX
#define GESUB cgesub
#define GEADD cgeadd
#else
#define GEADD sgeadd
#define GESUB sgesub
#endif /* CPLX */
#endif
#endif

#ifdef CPLX
#define SOPALIN_GEAM(i,j,m,n,x,a,u,b,v) GEADD((BLAS_FLOAT *) a,u,i,(BLAS_FLOAT *) b,v,j,(BLAS_FLOAT *) b,v,m,n)
#else
#define SOPALIN_GEAM(i,j,m,n,x,a,u,b,v) GEADD(a,u,i,b,v,j,b,v,m,n)
#endif /* CPLX */

#ifdef CPLX
#define SOPALIN_GESM(i,j,m,n,x,a,u,b,v) GESUB((BLAS_FLOAT *)b,v,j,(BLAS_FLOAT *)a,u,i,(BLAS_FLOAT *)b,v,m,n)
#else
#define SOPALIN_GESM(i,j,m,n,x,a,u,b,v) GESUB(b,v,j,a,u,i,b,v,m,n)
#endif


#ifdef BLAS_USE_COPY
#define SOPALIN_COPY(n,x,u,y,v) {memcpy(y,x,((INT)n)*sizeof(FLOAT));}
#else /* BLAS_USE_COPY */
#ifdef CPLX
#define SOPALIN_COPY(n,x,u,y,v) COPY(n,(BLAS_FLOAT *)x,u,(BLAS_FLOAT *)y,v)
#else
#define SOPALIN_COPY(n,x,u,y,v) COPY(n,x,u,y,v)
#endif
#endif /* BLAS_USE_COPY */

#define SOPALIN_LACPY(m, n, A, lda, B, ldb)                     \
  {                                                             \
    INT k;                                                      \
    for (k=0; k<(n); k++)                                       \
      {                                                         \
        SOPALIN_COPY(m, &(A[k*lda]), iun, &(B[k*ldb]), iun);    \
      }                                                         \
  }

#ifdef BLAS_USE_SCAL
#define SOPALIN_SCAL(n,a,x,u) {INT i;FLOAT *pt,*p=(x); \
  for(pt=p+n;p<pt;) *(p++)*=a; }
#else /* BLAS_USE_SCAL */
#ifdef CPLX
#define SOPALIN_SCAL(n,a,x,u) { FLOAT fgdg=a; SCAL(n,*((BLAS_FLOAT *) ((void *) &fgdg)),x,(BLAS_FLOAT *) u); }
#else
#define SOPALIN_SCAL(n,a,x,u) SCAL(n,a,x,u)
#endif
#endif /* BLAS_USE_SCAL */

#define SOPALIN_PPF(a,n,f)  PPF(a,n,f)

#ifdef CPLX
#define SOPALIN_TRSM(s,p,t,d,m,n,x,a,u,b,v) { FLOAT klsdfkls=x; TRSM(s,p,t,d,m,n,*((BLAS_FLOAT *) ((void *) &klsdfkls)),(BLAS_FLOAT *)a,u,(BLAS_FLOAT *)b,v); }
#else
#define SOPALIN_TRSM(s,p,t,d,m,n,x,a,u,b,v) TRSM(s,p,t,d,m,n,x,a,u,b,v)
#endif

#ifdef CPLX
#define SOPALIN_TRSV(p,t,d,n,a,u,b,v) { TRSV(p,t,d,n,(BLAS_FLOAT *)a,u,(BLAS_FLOAT *)b,v); }
#else
#define SOPALIN_TRSV(p,t,d,n,a,u,b,v) TRSV(p,t,d,n,a,u,b,v)
#endif

#define BLAS_GEMM
/*
#define IBM_GEMM
#define BLAS_GEMM
*/
#ifdef CPLX
#define SOPALIN_GEMM(i,j,m,n,k,x,a,u,b,v,y,c,w) { FLOAT fkldlgdkl=x,sdfmlgm=y; GEMM(i,j,m,n,k,*((BLAS_FLOAT *) ((void *) &fkldlgdkl)), (BLAS_FLOAT *) a,u,(BLAS_FLOAT *) b,v, *((BLAS_FLOAT *) ((void *) &sdfmlgm)),(BLAS_FLOAT *)c,w); }
#else
#define SOPALIN_GEMM(i,j,m,n,k,x,a,u,b,v,y,c,w) GEMM(i,j,m,n,k,x,a,u,b,v,y,c,w)
#endif

#ifdef CPLX
#define SOPALIN_GEMV(i,m,n,x,a,u,b,v,y,c,w) { FLOAT fghjk=x, fygtf=y; GEMV(i,m,n, *((BLAS_FLOAT *) ((void *) &fghjk)), (BLAS_FLOAT *) a,u,(BLAS_FLOAT *) b,v, *((BLAS_FLOAT *) ((void *) &fygtf)),(BLAS_FLOAT *) c,w); }
#define SOPALIN_AXPY(n,a,x,ix,y,iy) { FLOAT thjkl=a; AXPY(n,*((BLAS_FLOAT *) ((void *) &thjkl)),(BLAS_FLOAT *)x,ix,(BLAS_FLOAT *)y,iy); }
#else
#define SOPALIN_GEMV GEMV
#define SOPALIN_AXPY AXPY
#endif

#ifdef CPLX
#define SOPALIN_SYR(i,n,x,a,u,b,v) { FLOAT sgsdg=x; GER(n,n,*((BLAS_FLOAT *) ((void *) &sgsdg)),(BLAS_FLOAT *) a,u,(BLAS_FLOAT *) a,u,(BLAS_FLOAT *) b,v); }
#else
#define SOPALIN_SYR  SYR
#endif

#define SOPALIN_SYRK SYRK

#ifdef CPLX
#define SOPALIN_GER(m,n,x,a,u,b,v,c,w)  { FLOAT jkkd=x; GER(m,n,*((BLAS_FLOAT *) ((void *) &jkkd)),(BLAS_FLOAT *) a,u,(BLAS_FLOAT *) b,v,(BLAS_FLOAT *)c,w); }
#else
#define SOPALIN_GER  GER
#endif
#define SOPALIN_POF  POF
