/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/*
  File: sopalin_compute_linux.h

  Blas interface definition for LINUX/GNU operating systems.
 
  Sensible to those defines : 
    FORCE_DOUBLE   - Determine if coefficients are doubles or not.
    FORCE_COMPLEX  - Determine if coefficients are complex or not.
    INT            - Defined in common_pastix.h to determine integer type.
    CBLAS          - Determine if we want to use cblas or not.
    DEBUG_NAN      - If we want to check for NaN values in BLAS operartions, 
                     must be used with DBG_SOPALIN_NAN set to 1.
    BLAS_USE_COPY  - To use memcpy instead of BLAS copy function.
    BLAS_USE_SCAL  - To use "by hand" scal macro instead of BLAS one.

  Authors: 
    Ramet  Pierre  - ramet@labri.fr
    Xavier Lacoste - lacoste@labri.fr
*/

#ifndef X_INCLUDE_ESSL
/*#define CBLAS*/
#endif /* X_INCLUDE_ESSL */

/* 
   Define: BLAS_INT
   integer to use in blas calls.
*/
#define BLAS_INT INT

/*
  Define: BLAS macros

  Defines to set the correct blas depending on the
  coefficients type and the set of cblas usage or not.

*/
#ifdef FORCE_DOUBLE
#ifdef FORCE_COMPLEX
#ifdef CBLAS
/* -- Cblas with double complex coefficients -- */
#define AXPY cblas_zaxpy
#define GEMM cblas_zgemm
#define GEMV cblas_zgemv
#define TRSM cblas_ztrsm
#define TRSV cblas_ztrsv
#define GER  cblas_zgeru
#define SYR  cblas_zher
#define SYRK cblas_zherk
#define COPY cblas_zcopy
#define SCAL cblas_zscal
#else /* CBLAS */
/* -- Classical blas with double complex coefficients -- */
#define AXPY zaxpy_
#define GEMM zgemm_
#define GEMV zgemv_
#define TRSM ztrsm_
#define TRSV ztrsv_
#define POF  zpof_
#define PPF  zppf_
#define GER  zgeru_
#define SYR  zher_
#define SYRK zherk_
#define COPY zcopy_
#define SCAL zscal_
#endif /* CBLAS */
#else  /* CPLX */
#ifdef CBLAS
/* -- Cblas with double coefficients -- */
#define AXPY cblas_daxpy
#define GEMM cblas_dgemm
#define GEMV cblas_dgemv
#define TRSM cblas_dtrsm
#define TRSV cblas_dtrsv
#define GER  cblas_dger
#define SYR  cblas_dsyr
#define SYRK cblas_dsirk
#define COPY cblas_dcopy
#define SCAL cblas_dscal
#else /* CBLAS */
/* Classical blas with double coefficients         */
#define AXPY                        daxpy_
#define GEMM                        dgemm_
#define GEMV                        dgemv_
#define GEAM                        dgeam_
#define GECP                        dgecp_
#define SYMM                        gsymm_
#define TRSM                        dtrsm_
#define TRSV                        dtrsv_
#define GEF                         dgef_
#define POF                         dpof_
#define PPF                         dppf_
#define GER                         dger_
#define GETMO                       dgetmo_
#define SYR                         dsyr_
#define SPR                         dspr_
#define SYRK                        dsyrk_
#define COPY                        dcopy_
#define SCAL                        dscal_
#define V_DIV                       dv_div_
#define POTRF                       dpotrf_
#define POTF2                       dpotf2_
#endif /* CBLAS */
#endif /* CPLX */
#else /* BLAS_DOUBLE */
#ifdef CPLX
#ifdef CBLAS
/* -- Cblas with simple complex coefficients -- */
#define AXPY cblas_caxpy
#define GEMM cblas_cgemm
#define GEMV cblas_cgemv
#define TRSM cblas_ctrsm
#define TRSV cblas_ctrsv
#define GER  cblas_cgeru
#define SYR  cblas_cher
#define SYRK cblas_cherk
#define COPY cblas_ccopy
#define SCAL cblas_cscal
#else /* CBLAS */
/* -- Classical blas with simple complex coefficients -- */
#define AXPY caxpy_
#define GEMM cgemm_
#define GEMV cgemv_
#define TRSM ctrsm_
#define TRSV ctrsv_
#define GER  cgeru_
#define SYR  cher_
#define SYRK cherk_
#define COPY ccopy_
#define SCAL cscal_
#endif /* CBLAS */
#else /* CPLX */
#ifdef CBLAS
/* -- Cblas with simple coefficients -- */
#define AXPY cblas_saxpy
#define GEMM cblas_sgemm
#define GEMV cblas_sgemv
#define TRSM cblas_strsm
#define TRSV cblas_strsv
#define GER  cblas_sger
#define SYR  cblas_ssyr
#define SYRK cblas_ssyrk
#define COPY cblas_scopy
#define SCAL cblas_sscal
#else /* CBLAS */
/* -- Classical blas with simple coefficients -- */
#define AXPY                        saxpy_
#define GEMM                        sgemm_
#define GEMV                        sgemv_
#define GEAM                        sgeam_
#define GECP                        sgecp_
#define SYMM                        ssymm_
#define TRSM                        strsm_
#define TRSV                        strsv_
#define GEF                         sgef_
#define POF                         spof_
#define PPF                         sppf_
#define GER                         sger_
#define GETMO                       sgetmo_
#define SYR                         ssyr_
#define SPR                         sspr_
#define SYRK                        ssyrk_
#define COPY                        scopy_
#define SCAL                        sscal_
#define V_DIV                       sv_div_
#define POTRF                       spotrf_
#define POTF2                       spotf2_
#endif /* CBLAS */
#endif /* CPLX */
#endif

void dim_dgeam(char *transa,char *transb,INT m,INT n,FLOAT alpha,FLOAT *a, INT lda,FLOAT *b,INT ldb);

/* Pb pas de truc d'addition C=A+B */
/*
  Macro: SOPALIN_GEAM

  Computes $ b = x \times a + b $.

  Parameters: 
     i - indicates if a needs to be transposed.
     j - indicates if b needs to be transposed.
     m - number of row in a and b. 
     n - number of colonnes in a and b.
     x - scalar.
     a - Matrix a.
     u - Stride between 2 columns of a.
     b - Matrix b.
     v - Stride between 2 columns of b.   
*/
#ifdef DEBUG_NAN
#define SOPALIN_GEAM(i,j,m,n,x,a,u,b,v) {	\
    INT iter, iter2;				\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)a)[iter2+iter*u]);	\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter2+iter*v]);	\
    dim_dgeam(i,j,m,n,x,a,u,b,v);		\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter*v+iter2]);	\
  }
#else
#define SOPALIN_GEAM(i,j,m,n,x,a,u,b,v) dim_dgeam(i,j,m,n,x,a,u,b,v)
#endif
/* Pb pas de truc de soustraction C=A-B *//*
  Macro: SOPALIN_GEAM

  Computes $ b = x \times a - b $.

  Parameters: 
     i - indicates if a needs to be transposed.
     j - indicates if b needs to be transposed.
     m - number of row in a and b. 
     n - number of colonnes in a and b.
     x - scalar.
     a - Matrix a.
     u - Stride between 2 columns of a.
     b - Matrix b.
     v - Stride between 2 columns of b.   
*/
#ifdef DEBUG_NAN
#define SOPALIN_GESM(i,j,m,n,x,a,u,b,v) {	\
    INT iter, iter2;				\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)a)[iter2+iter*u]);	\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter2+iter*v]);	\
    dim_dgeam(i,j,m,n,-(x),a,u,b,v);		\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter*v+iter2]);	\
  }
#else
#define SOPALIN_GESM(i,j,m,n,x,a,u,b,v) dim_dgeam(i,j,m,n,-(x),a,u,b,v)
#endif

/* 
   Macro: SOPALIN_COPY

   Copy *x* vector to *y* vector.

   Care: if BLAS_USE_COPY is defined, only works
     with *u* and *v* equal to 1.

   Parameters:
     n - Size of the vectors.
     x - Source vector
     u - *x* vector stride between to cefficients.
     y - Destination vector.
     v - *y* vector stride between to cefficients.
*/
#ifdef BLAS_USE_COPY
#define SOPALIN_COPY(n,x,u,y,v) {		\
    memcpy((void *)(y), (void *)(x),		\
	   ((INT)(n))*sizeof(FLOAT));		\
  }
#else /* BLAS_USE_COPY */
extern void COPY(BLAS_INT *n, FLOAT *x, BLAS_INT * u, FLOAT * y, BLAS_INT * v);

#ifdef DEBUG_NAN
#define SOPALIN_COPY(n,x,u,y,v) {		\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    BLAS_INT variv = (BLAS_INT)(v);		\
    INT iter;					\
    for (iter = 0; iter < n; iter++)		\
      CHECK_NAN(((FLOAT*)x)[iter]);		\
    for (iter = 0; iter < n; iter++)		\
      CHECK_NAN(((FLOAT*)y)[iter]);		\
    COPY(&varin, (x), &variu, (y), &variv);	\
    for (iter = 0; iter < n; iter++)		\
      CHECK_NAN(((FLOAT*)y)[iter]);		\
  }
#else
#define SOPALIN_COPY(n,x,u,y,v) {		\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    BLAS_INT variv = (BLAS_INT)(v);		\
    COPY(&varin, (x), &variu, (y), &variv);	\
  }
#endif
#endif /* BLAS_USE_COPY */

/* 
   Macro: SOPALIN_LACPY

   Equivalent of lacpy form LAPACK with uplo = 'General' 

   Care: if BLAS_USE_COPY is defined, only works
     with *lda* and *ldb* equal to 1.

   Parameters:
     m   - Number of rows to copy
     n   - Number of columns to copy
     A   - lda-by-n Matrix A 
     lda - Leading dimension of A (lda >= m)
     B   - Destination ldb-by-n matrix B
     ldb - Leading dimension of B (ldb >= m)

*/
#define SOPALIN_LACPY(m, n, A, lda, B, ldb)                         \
  {                                                                 \
    INT k;                                                          \
    for (k=0; k<(n); k++)                                           \
      {                                                             \
        SOPALIN_COPY(m, &(A[k*(lda)]), iun, &(B[k*(ldb)]), iun);    \
      }                                                             \
  }


/*
  Macro: SOPALIN_SCAL

  Multiply a vector by a scalar.
  
  Care: if BLAS_USE_SCAL is defined, only works
    with *u* equal to 1.

  Parameters:
    n - Size of the vector
    a - Scalar to multiply the vector by.
    x - Vector.
    u - stride between to element of the vector.

*/
#ifdef BLAS_USE_SCAL
#define SOPALIN_SCAL(n,a,x,u) {			\
    INT i; FLOAT *pt,*p=(x);			\
    for((pt)=(p+n);(p)<(pt);)			\
      *((p)++)*= (a);				\
  }
#else /* BLAS_USE_SCAL */
extern void SCAL(BLAS_INT * n , FLOAT * a, FLOAT * x, BLAS_INT * u);
#ifdef DEBUG_NAN
#define SOPALIN_SCAL(n,a,x,u) {			\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    FLOAT    varia = (FLOAT)(a);		\
    INT iter;					\
    for (iter = 0; iter < n; iter++)		\
      CHECK_NAN(((FLOAT*)x)[iter]);		\
    SCAL(&varin, &varia, (x), &variu);		\
    for (iter = 0; iter < n; iter++)		\
      CHECK_NAN(((FLOAT*)x)[iter]);		\
  }
#else
#define SOPALIN_SCAL(n,a,x,u) {			\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    FLOAT    varia = (FLOAT)(a);		\
    SCAL(&varin, &varia, (x), &variu); }
#endif
#endif /* BLAS_USE_SCAL */

extern void TRSM (char * s, char * p, char * t, char * d, BLAS_INT * m, 
		  BLAS_INT * n , FLOAT * x, FLOAT * a, BLAS_INT * u, 
		  FLOAT * b, BLAS_INT * v);
/*
  Macro: SOPALIN_TRSM
  
  Solves $ A  \times  X = x  \times  B $ or $ X  \times  A = x  \times  B $ depending on 
  s value.
  A is a triangular matrix.

  Parameters:
    s - Position of *a* 'L' or 'l' for left, 'R' or 'r' for right.
    p - Indicate if the matrice is upper ('U') or lower ('L') 
        triangular matrix.
    t - Indicate if user wants to transpose *a*.
    d - Indicate if diagonal is united ('U') or not ('N').
    m - Number of lines in *b*.
    n - Number of columns in *b*.
    x - Coefficient to multiply right-hand-side by.
    a - *a* matrix (MxM or NxN depending on s).
    u - Stride for *a*
    b - *b* MxN matrix.
    v - Stride for *b*.
*/
#ifdef DEBUG_NAN
#define SOPALIN_TRSM(s,p,t,d,m,n,x,a,u,b,v) {	\
    BLAS_INT varim = (BLAS_INT)(m);		\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    BLAS_INT variv = (BLAS_INT)(v);		\
    FLOAT    varix = (FLOAT)(x);		\
    INT iter, iter2;				\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < m; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter*v+iter2]);	\
    if ( *s == 'L' || *s == 'l' ) {		\
      for (iter = 0; iter < m; iter++)		\
	for (iter2 = 0; iter2 < m; iter2++)	\
	  CHECK_NAN(((FLOAT*)a)[iter*u+iter2]);	\
    }						\
    else {					\
      for (iter = 0; iter < n; iter++)		\
	for (iter2 = 0; iter2 < n; iter2++)	\
	  CHECK_NAN(((FLOAT*)a)[iter*u+iter2]);	\
    }						\
    TRSM((s), (p), (t), (d), &varim, &varin,	\
	 &varix, (a), &variu, (b), &variv);	\
						\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < m; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter*v+iter2]);	\
  }
#else
#define SOPALIN_TRSM(s,p,t,d,m,n,x,a,u,b,v) {				\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    FLOAT    varix = (FLOAT)(x);					\
    TRSM((s), (p), (t), (d), &varim, &varin, &varix, (a), &variu, (b), &variv); }
#endif

extern void TRSV(char *, char *, char *, BLAS_INT *, FLOAT *, BLAS_INT *, FLOAT *, BLAS_INT *);
#ifdef DEBUG_NAN
/*
  Macro: SOPALIN_TRSV

  Solves a system of linear equation with 
  a triangular matrix.

  Parameters:
    p - Indicate if the matrix is upper ('U') or lower ('L') triangular.
    t - Indicate if user wants to transpose ('T') the matrix or not ('N').
    d - Indicate if the matrix has a united diagonal ('U') ot not ('N').
    n - Size of the matrix and vector.
    a - Triangular matrix NxN.
    u - Stride for *a*.
    b - Right-hand-side member.
    v - Stride for *b*.
*/
#define SOPALIN_TRSV(p,t,d,n,a,u,b,v) {				\
    BLAS_INT varin = (BLAS_INT)(n);				\
    BLAS_INT variu = (BLAS_INT)(u);				\
    BLAS_INT variv = (BLAS_INT)(v);				\
    INT iter, iter2;						\
    for (iter = 0; iter < n; iter++)				\
      for (iter2 = 0; iter2 < n; iter2++)			\
	CHECK_NAN(((FLOAT*)a)[iter*u*iter2]);			\
    for (iter = 0; iter < n; iter++)				\
      CHECK_NAN(((FLOAT*)b)[iter]);				\
								\
    TRSV((p), (t), (d), &varin, (a), &variu, (b), &variv);	\
								\
    for (iter = 0; iter < n; iter++)				\
      CHECK_NAN(((FLOAT*)b)[iter]);				\
  }
#else
#define SOPALIN_TRSV(p,t,d,n,a,u,b,v) {				\
    BLAS_INT varin = (BLAS_INT)(n);				\
    BLAS_INT variu = (BLAS_INT)(u);				\
    BLAS_INT variv = (BLAS_INT)(v);				\
								\
    TRSV((p), (t), (d), &varin, (a), &variu, (b), &variv);	\
								\
  }
#endif

extern void GEMM(char * i, char * j, BLAS_INT * m, BLAS_INT * n, 
		 BLAS_INT * k, FLOAT * x, FLOAT * a, BLAS_INT * u, 
		 FLOAT * b, BLAS_INT * v, FLOAT * y, FLOAT * c, BLAS_INT * w);
/*
  Macro: SOPALIN_GEMM

  Computes $ c = x \times a \times b + y \times c $.

  Parameters:
    i - indicate if user wants to transpose *a*.
    j - indicate if user wants to transpose *b*.
    m - Number of lines in *c* and *a*.
    n - Number of columns in *c* and *b*.
    k - Number of lines in *b* and columns in *a*.
    x - coefficient for $ a \times b$ 
    a - $m \times k$ Matrix.
    u - Stride for *a*.
    b - $k \times n$ Matrix.
    v - Stride for *b*.
    y - Coefficient for *c*
    c - $m \times n$ matrix.
    w - Stride for *c*
*/
#ifdef DEBUG_NAN
#define SOPALIN_GEMM(i,j,m,n,k,x,a,u,b,v,y,c,w) {			\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT varik = (BLAS_INT)(k);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    BLAS_INT variw = (BLAS_INT)(w);					\
    FLOAT    varix = (FLOAT)(x);					\
    FLOAT    variy = (FLOAT)(y);					\
    INT iter, iter2;							\
    for(iter = 0; iter < k; iter ++)					\
      for(iter2 = 0; iter2 < m; iter2 ++)				\
	CHECK_NAN(((FLOAT*)a)[iter*u+iter2]);				\
    for(iter = 0; iter < n; iter ++)					\
      for(iter2 = 0; iter2 < k; iter2 ++)				\
	CHECK_NAN(((FLOAT*)b)[iter*u+iter2]);				\
    for(iter = 0; iter < n; iter ++)					\
      for(iter2 = 0; iter2 < m; iter2 ++)				\
	CHECK_NAN(((FLOAT*)c)[iter2+iter*u]);				\
									\
    GEMM((i), (j), &varim, &varin, &varik, &varix, (a), &variu, (b),	\
	 &variv, &variy, (c), &variw);					\
									\
    for(iter = 0; iter < n; iter ++)					\
      for(iter2 = 0; iter2 < m; iter2 ++)				\
	CHECK_NAN(((FLOAT*)c)[iter*u+iter2]);				\
  }
#else
#define SOPALIN_GEMM(i,j,m,n,k,x,a,u,b,v,y,c,w) {			\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT varik = (BLAS_INT)(k);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    BLAS_INT variw = (BLAS_INT)(w);					\
    FLOAT    varix = (FLOAT)(x);					\
    FLOAT    variy = (FLOAT)(y);					\
    GEMM((i), (j), &varim, &varin, &varik, &varix, (a), &variu, (b), &variv, &variy, (c), &variw); }
#endif
extern void GEMV(char *, BLAS_INT *, BLAS_INT *, FLOAT*, FLOAT*, BLAS_INT*, FLOAT*,BLAS_INT*,FLOAT*,FLOAT*,BLAS_INT*);
#ifdef DEBUG_NAN
#define SOPALIN_GEMV(i,m,n,x,a,u,b,v,y,c,w) {		\
    BLAS_INT varim = (BLAS_INT)(m);			\
    BLAS_INT varin = (BLAS_INT)(n);			\
    BLAS_INT variu = (BLAS_INT)(u);			\
    BLAS_INT variv = (BLAS_INT)(v);			\
    BLAS_INT variw = (BLAS_INT)(w);			\
    FLOAT    varix = (FLOAT)(x);			\
    FLOAT    variy = (FLOAT)(y);			\
    INT iter, iter2;					\
							\
    for (iter = 0; iter < n; iter ++)			\
      for (iter2 = 0; iter2 < m; iter2++)		\
	CHECK_NAN(((FLOAT*)a)[iter2+iter*u]);		\
    if (*i == 'N' || *i == 'n')	{			\
      for (iter = 0; iter < n; iter++)			\
	CHECK_NAN(((FLOAT*)b)[iter*v]);			\
      for (iter = 0; iter < m; iter++)			\
	CHECK_NAN(((FLOAT*)c)[iter*w]);			\
    }							\
    else {						\
      for (iter = 0; iter < m; iter++)			\
	CHECK_NAN(((FLOAT*)b)[iter*v]);			\
      for (iter = 0; iter < n; iter++)			\
	CHECK_NAN(((FLOAT*)c)[iter*w]);			\
    }							\
    GEMV((i), &varim, &varin, &varix, (a), &variu,	\
	 (b), &variv, &variy, (c), &variw);		\
    if (*i == 'N' || *i == 'n')	{			\
      for (iter = 0; iter < m; iter++)			\
	CHECK_NAN(((FLOAT*)c)[iter*w]);			\
    }							\
    else {						\
      for (iter = 0; iter < n; iter++)			\
	CHECK_NAN(((FLOAT*)c)[iter*w]);			\
    }							\
  }

#else
#define SOPALIN_GEMV(i,m,n,x,a,u,b,v,y,c,w) {				\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    BLAS_INT variw = (BLAS_INT)(w);					\
    FLOAT    varix = (FLOAT)(x);					\
    FLOAT    variy = (FLOAT)(y);					\
    GEMV((i), &varim, &varin, &varix, (a), &variu, (b), &variv, &variy, (c), &variw); }
#endif

extern void AXPY(BLAS_INT * n , FLOAT * a, FLOAT * x, BLAS_INT *  ix, FLOAT * y, BLAS_INT * iy);

#ifdef DEBUG_NAN
#define SOPALIN_AXPY(n,a,x,ix,y,iy) {			\
    BLAS_INT varin = (BLAS_INT)(n);			\
    BLAS_INT varix = (BLAS_INT)(ix);			\
    BLAS_INT variy = (BLAS_INT)(iy);			\
    FLOAT    varia = (FLOAT)(a);			\
    INT iter;						\
    for (iter = 0; iter < n; iter ++)			\
      CHECK_NAN(((FLOAT*)x)[iter*ix]);			\
    for (iter = 0; iter < n; iter ++)			\
      CHECK_NAN(((FLOAT*)y)[iter*iy]);			\
    AXPY(&varin, &varia, (x), &varix, (y), &variy);	\
    for (iter = 0; iter < n; iter ++)			\
      CHECK_NAN(((FLOAT*)y)[iy*iter]);			\
  }
#else
#define SOPALIN_AXPY(n,a,x,ix,y,iy) {			\
    BLAS_INT varin = (BLAS_INT)(n);			\
    BLAS_INT varix = (BLAS_INT)(ix);			\
    BLAS_INT variy = (BLAS_INT)(iy);			\
    FLOAT    varia = (FLOAT)(a);			\
    AXPY(&varin, &varia, (x), &varix, (y), &variy); }
#endif

#ifdef CPLX
#ifdef DEBUG_NAN
#define SOPALIN_SYR(i,n,x,a,u,b,v) {					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    FLOAT    varix = (FLOAT)(x);					\
    INT      iter, iter2;						\
    for (iter = 0; iter < n; iter++)					\
      for (iter2 = 0; iter2 < n; iter2++)				\
	CHECK_NAN(((FLOAT*)b)[iter*u+iter2]);				\
    for (iter = 0; iter < n; iter++)					\
      CHECK_NAN(((FLOAT*)a)[iter]);					\
    GER(&varin, &varin, &varix, (a), &variu, (a), &variu, (b), &variv);	\
    for (iter = 0; iter < n; iter++)					\
      for (iter2 = 0; iter2 < n; iter2++)				\
	CHECK_NAN(((FLOAT*)b)[iter2+iter*u]);				\
  }
#else /* DEBUG_NAN */
#define SOPALIN_SYR(i,n,x,a,u,b,v) {					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    FLOAT    varix = (FLOAT)(x);					\
    GER(&varin, &varin, &varix, (a), &variu, (a), &variu, (b), &variv); }
#endif /* DEBUG_NAN */
#else  /* CPLX */
extern void SYR(char * i, BLAS_INT * n , FLOAT * x, FLOAT * a, BLAS_INT * u, FLOAT * b, BLAS_INT * v);
#ifdef DEBUG_NAN
#define SOPALIN_SYR(i,n,x,a,u,b,v) {			\
    BLAS_INT varin = (BLAS_INT)(n);			\
    BLAS_INT variu = (BLAS_INT)(u);			\
    BLAS_INT variv = (BLAS_INT)(v);			\
    FLOAT    varix = (FLOAT)(x);			\
    INT iter, iter2;					\
    for (iter = 0; iter < n; iter++)			\
      for (iter2 = 0; iter2 < n; iter2++)		\
	CHECK_NAN(((FLOAT*)b)[iter*u+iter2]);		\
    for (iter = 0; iter < n; iter++)			\
      CHECK_NAN(((FLOAT*)a)[iter]);			\
    SYR((i), &varin, &varix, (a), &variu, (b), &variv);	\
    for (iter = 0; iter < n; iter++)			\
      for (iter2 = 0; iter2 < n; iter2++)		\
	CHECK_NAN(((FLOAT*)b)[iter2+iter*u]);		\
  }
#else
#define SOPALIN_SYR(i,n,x,a,u,b,v) {				\
    BLAS_INT varin = (BLAS_INT)(n);				\
    BLAS_INT variu = (BLAS_INT)(u);				\
    BLAS_INT variv = (BLAS_INT)(v);				\
    FLOAT    varix = (FLOAT)(x);				\
    SYR((i), &varin, &varix, (a), &variu, (b), &variv); }
#endif
#endif
extern void SYRK (char * i, char * t,  BLAS_INT * n, BLAS_INT * k, FLOAT * x, FLOAT * a, BLAS_INT * u, FLOAT * y, FLOAT * b, BLAS_INT *v);
#ifdef DEBUG_NAN
#define SOPALIN_SYRK(i,t,n,k,x,a,u,y,b,v) {	\
    BLAS_INT varin = (BLAS_INT)(n);		\
    BLAS_INT varik = (BLAS_INT)(k);		\
    BLAS_INT variu = (BLAS_INT)(u);		\
    BLAS_INT variv = (BLAS_INT)(v);		\
    FLOAT    varix = (FLOAT)(x);		\
    FLOAT    variy = (FLOAT)(y);		\
    INT iter, iter2;				\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[iter2+v*iter]);	\
    if (*t == 'N' || *t == 'N') {		\
      for (iter = 0; iter < k; iter++)		\
	for (iter2 = 0; iter2 < n; iter2++)	\
	  CHECK_NAN(((FLOAT*)a)[iter2+v*iter]);	\
    }						\
    else {					\
      for (iter = 0; iter < n; iter++)		\
	for (iter2 = 0; iter2 < k; iter2++)	\
	  CHECK_NAN(((FLOAT*)a)[iter2+v*iter]);	\
    }						\
    SYRK((i), (t), &varin, &varik, &varix, (a),	\
	 &variu, &variy, (b), &variv);		\
    for (iter = 0; iter < n; iter++)		\
      for (iter2 = 0; iter2 < n; iter2++)	\
	CHECK_NAN(((FLOAT*)b)[v*iter+iter2]);	\
  }
#else
#define SOPALIN_SYRK(i,t,n,k,x,a,u,y,b,v) {				\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT varik = (BLAS_INT)(k);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    FLOAT    varix = (FLOAT)(x);					\
    FLOAT    variy = (FLOAT)(y);					\
    SYRK((i), (t), &varin, &varik, &varix, (a),				\
	 &variu, &variy, (b), &variv);					\
  }
#endif
extern void GER(BLAS_INT * m, BLAS_INT * n, FLOAT * x, FLOAT * a, BLAS_INT * u, FLOAT * b, BLAS_INT * v, FLOAT * c, BLAS_INT * w );
#ifdef DEBUG_NAN
#define SOPALIN_GER(m,n,x,a,u,b,v,c,w) {				\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    BLAS_INT variw = (BLAS_INT)(w);					\
    FLOAT    varix = (FLOAT)(x);					\
    INT iter, iter2;							\
    for (iter = 0; iter < n; iter++) {					\
      for (iter2 = 0; iter2 < m; iter2++) {				\
	CHECK_NAN(((FLOAT*)c)[iter*w+iter2]);				\
      }									\
    }									\
    for (iter = 0; iter < n; iter++) {					\
      CHECK_NAN(((FLOAT*)a)[iter*u]);					\
    }									\
    for (iter = 0; iter < n; iter++) {					\
      CHECK_NAN(((FLOAT*)b)[iter*v]);					\
    }									\
    GER(&varim, &varin, &varix, (a), &variu, (b), &variv, (c), &variw); \
    for (iter = 0; iter < n; iter++)					\
      for (iter2 = 0; iter2 < m; iter2++)				\
	CHECK_NAN(((FLOAT*)c)[iter2+iter*w]);				\
  }
#else
#define SOPALIN_GER(m,n,x,a,u,b,v,c,w) {				\
    BLAS_INT varim = (BLAS_INT)(m);					\
    BLAS_INT varin = (BLAS_INT)(n);					\
    BLAS_INT variu = (BLAS_INT)(u);					\
    BLAS_INT variv = (BLAS_INT)(v);					\
    BLAS_INT variw = (BLAS_INT)(w);					\
    FLOAT    varix = (FLOAT)(x);					\
    GER(&varim, &varin, &varix, (a), &variu, (b), &variv, (c), &variw); \
  }
#endif

#define SOPALIN_POF(i,a,u,n)  POF((i), (a), (u), (n))
#define SOPALIN_PPF(a,n,f)    PPF((a), (n), (f))

